/* ftpq: ftp.c

Copyright (C) 2000, 2001 Mark Tranchant

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "ftpq.h"


/* Make the FTP data connection, using the parameters returned by the
   PASV command (IP address, port high/low bytes). Return the socket
   descriptor. */

int data_connect (int a1, int a2, int a3, int a4, int p1, int p2) {

	int sockfd;
	char name[16];
	int port;

	sprintf (name, "%d.%d.%d.%d", a1, a2, a3, a4);
	port = (p1 * 256) + p2;

	sockfd = get_socket_out (name, port);

	return sockfd;
}


/* Connect to the named server/port, get the welcome message, and return
   the socket descriptor. */

int svr_login (char *name, int port) {

	int sockfd;

	sockfd = get_socket_out (name, port);

	if (sockfd != 0)
		get_response (sockfd);
  
	return sockfd;
}


/* Send a command to the FTP/SMTP server. If verbose mode is specified,
   the sent command is printed: in the case of the PASS command, the password
   is replaced with asterisks. If a reply is expected, get_response() (below)
   is called, and its return value is returned to our caller. */

int send_cmd (char *cmd, int sockfd, int resp) {

	int ctr, stars;

	if (cmd != NULL) {
		send (sockfd, cmd, strlen (cmd), 0);

		if (gv.verbose == VERBOSE) {
			if ((char *) strstr (cmd, "PASS ") == cmd) {
				printf ("-> PASS ");
				stars = strlen (cmd) - 6;
				for (ctr = 0; ctr < stars; ctr++) {
					printf ("*");
				}
				printf ("\n");
			}
			else
				printf ("-> %s", cmd);
		}
	}

	if (resp == NORESP)
		return OK;
	else
		return get_response (sockfd);
}


/* Read the response from the FTP server, put the last line into the global
   hostresp[] array (statically defined in ftpq.h) and return the response
   code. Servers can return multiple-line responses as multiple messages
   or as a single message including line breaks. See RFC959 for details of how
   the end of the message is denoted.
   This also works for SMTP responses, and is used by mailer.c */

int get_response (int sockfd) {

	int charctr = 0;
	char retcode[4];
	int firstcode = 0;
	enum {D2, D3, SEP, NL, WAIT, LAST, DONE} state = NL;
	int numbytes;

	retcode[3] = '\0';
	memset (hostresp, '\0', HOSTRESP);

	while (state != DONE) {
		if ((numbytes = recvfrom (sockfd, hostresp, HOSTRESP, 0, NULL, 0)) == -1)
			return 600;
		hostresp[numbytes] = '\0';
		charctr = 0;
		while (charctr < numbytes) {
			switch (state) {
			  case NL:
				if (gv.verbose == VERBOSE)
					printf ("<- ");
				if (hostresp[charctr] == '\n')
					state = NL;
				else if (isdigit ((int) hostresp[charctr])) {
					retcode[0] = hostresp[charctr];
					state = D2;
				}
				else
					state = WAIT;
				break;
			  case D2:
				if (isdigit ((int) hostresp[charctr])) {
					retcode[1] = hostresp[charctr];
					state = D3;
				}
				else if (hostresp[charctr] == '\n')
					state = NL;
				else
					state = WAIT;
				break;
			  case D3:
				if (isdigit ((int) hostresp[charctr])) {
					retcode[2] = hostresp[charctr];
					state = SEP;
				}
				else if (hostresp[charctr] == '\n')
					state = NL;
				else
					state = WAIT;
				break;
			  case SEP:
				if (hostresp[charctr] == ' ') {
					if (firstcode == 0) {
						firstcode = atoi (retcode);
						state = LAST;
					}
					else if (firstcode != atoi (retcode))
						state = WAIT;
					else
						state = LAST;
				}
				else if (hostresp[charctr] == '-' && firstcode == 0) {
					firstcode = atoi (retcode);
					state = WAIT;
				}
				else if (hostresp[charctr] == '\n')
					state = NL;
				else
					state = WAIT;
				break;
			  case WAIT:
				if (hostresp[charctr] == '\n')
					state = NL;
				break;
			  case LAST:
				if (hostresp[charctr] == '\n')
					state = DONE;
				break;
			  case DONE:
				break;
			  default:
				printf ("Ouch - ftp state machine error!\n");
				exit (AB_CASEERR);
			}
			if (gv.verbose == VERBOSE)
				printf ("%c", hostresp[charctr]);
			charctr++;
		}
	}

	fflush (stdout);

	return firstcode;
}


/* category (response) categorizes an FTP response into one of the
   five response "types" listed in RFC959, or a default error category. */

int category (int response) {

	int reply = RESP_ERROR;

	if (response >= 600)
		reply = RESP_ERROR;
	else if (response >= 500)
		reply = NEG_PERM;
	else if (response >= 400)
		reply = NEG_TRANS;
	else if (response >= 300)
		reply = POS_INTER;
	else if (response >= 200)
		reply = POS_COMPLETE;
	else if (response >= 100)
		reply = POS_PRELIM;
	else
		reply = RESP_ERROR;

	return reply;
}


int svr_logout (int sockfd) {

        int response;
 
        response = ftp_quit (sockfd);

        return OK;
}

int usr_login (char *udir, char *uname, int sockfd) {

	int response;
	char *password = NULL;
	char *account = NULL;
	char line [1024];
	char buf [1024]; 
        FILE *usrptr = NULL;
        char *usrfile = NULL;

        usrfile = memalloc (strlen (udir) + 7);
        sprintf (usrfile, "%s/user", udir);
        usrptr = fileopen (usrfile, FO_READ | FO_NEED);

	if (usrptr == NULL)
		return OK;

	while (fgets (line, 1023, usrptr) != NULL) {	
		if (sscanf (line, "p: %s", buf)) {
			password = memalloc (strlen (buf) + 1);
			strcpy (password, buf);
		}
		else if (sscanf (line, "a: %s", buf)) {
			account = memalloc (strlen (buf) + 1);
			strcpy (account, buf);
		}
	}

	fclose (usrptr);
	free (usrfile);

	response = ftp_user (uname, sockfd);

	if (response == 331) { 
		response = ftp_pass (password, sockfd);
		free (password);
	}
	if (response == 332) {
		response = ftp_acct (account, sockfd);
		free (account);	
	}

	response = ftp_typei (sockfd);
	response = ftp_modes (sockfd);
	response = ftp_struf (sockfd);

	return OK;

}

int send_data (int datafd) {

	FILE *this_file = NULL;
	long totalgone, bytesize, gone;
	int tosend;
	struct stat filedet;
	struct timeval *tv;
	long now, mnow, then, mthen, elapsed, rate;

	stat (gv.srcpath, &filedet);
	bytesize = (off_t) filedet.st_size;

	this_file = fileopen (gv.srcpath, FO_READ | FO_NONEED);

	if (this_file == NULL)
		return -1;

	totalgone = 0L;

	tv = memalloc (sizeof (struct timeval));
	gettimeofday (tv, NULL);
	then = tv->tv_sec;
	mthen = tv->tv_usec >> 10;

	while (totalgone < bytesize) {
		tosend = fread (buffer, 1, (size_t) BUFSIZE, this_file);
		gone = 0L;
		while (gone < tosend) {
			gone += sendto (datafd, buffer + gone, tosend - gone, 0, NULL, 0);
			totalgone += gone;
		}
		if (gv.rate > 0) {
			gettimeofday (tv, NULL);
			now = tv->tv_sec;
			mnow = tv->tv_usec >> 10;
			elapsed = ((now - then) << 10) + (mnow - mthen);
			if (elapsed > 0L) {
				while ((rate = ((totalgone << 10) / elapsed)) > gv.rate)  {
					gettimeofday (tv, NULL);
					now = tv->tv_sec;
					mnow = tv->tv_usec >> 10; 
					elapsed = ((now - then) << 10) + (mnow - mthen);
#ifdef HAVE_USLEEP
					(void) usleep ((unsigned long) 200000L);
#endif
#ifndef HAVE_USLEEP
#ifdef HAVE_SLEEP
					sleep (1);
#endif
#endif

				}
			}
		}
	}

	gettimeofday (tv, NULL);
	now = tv->tv_sec;
	mnow = tv->tv_usec >> 10;
	elapsed = ((now - then) << 10) + (mnow - mthen);

	gv.transfer += totalgone;
	gv.elapsed += elapsed;

	free (tv);
	fclose (this_file);
	return 0;
}

int get_data (int datafd) {

	FILE *this_file = NULL;
	long totalgot, got;

	this_file = fileopen (gv.dstpath, FO_WRITE | FO_NEED);

	if (this_file == NULL)
		return -1;

	totalgot = 0;
	
	while ((got = recvfrom (datafd, buffer, (size_t) BUFSIZE, 0, NULL, 0)) > 0) {
		fwrite (buffer, 1, got, this_file);
		totalgot += got;
	}

	fclose (this_file);
	return 0;
}

